from pathlib import Path
from typing import NamedTuple


class RasterBaseCoordinates(NamedTuple):
    """Base coordinates for a SRTM HGT raster file

    HGT files are named based on the coordinate of their lower-left corner
    """

    latitude: int
    longitude: int

    @classmethod
    def from_float(cls, latitude: float, longitude: float):
        # Raster base coordinates are measured from the bottom left (i.e SE)
        # corner. So we need calculate the base coordate a bit differently for
        # negative lat/longs.
        if latitude < 0:
            latitude -= 1
        if longitude < 0:
            longitude -= 1

        return cls(latitude=int(latitude), longitude=int(longitude))

    @classmethod
    def from_file_name(cls, hgt_name: str):
        """Create an instance from a HGT name, eg "N38W006" """
        hgt_name = hgt_name.upper()
        error = f"Invalid hgt name ({hgt_name}), expected format (N|S)00(E|W)000"
        assert len(hgt_name) == 7, error
        assert hgt_name[0] in ("N", "S"), error
        assert hgt_name[3] in ("E", "W"), error
        try:
            is_north = hgt_name[0] == "N"
            latitude = int(hgt_name[1:3])
            is_east = hgt_name[3] == "E"
            longitude = int(hgt_name[4:7])
        except ValueError:
            assert False, error

        if not is_north:
            latitude *= -1
        if not is_east:
            longitude *= -1

        return cls(latitude=latitude, longitude=longitude)

    @classmethod
    def from_file_path(cls, path: Path):
        """Create an instance from a path to a HGT file"""
        return cls.from_file_name(hgt_name=path.name.split(".")[0])

    @property
    def file_name(self):
        """Get the expected HGT file name for these base coordinates (without file extension)"""
        if self.latitude >= 0:
            latitude = f"N{self.latitude:0>2}"
        else:
            latitude = f"S{-self.latitude:0>2}"

        if self.longitude >= 0:
            longitude = f"E{self.longitude:0>3}"
        else:
            longitude = f"W{-self.longitude:0>3}"

        return f"{latitude}{longitude}"



from math import sin, cos, radians, atan2, sqrt, ceil
import os
from pathlib import Path
from statistics import mean
from typing import List, Tuple, NamedTuple

EARTH_RADIUS = 6373000
METERS_PER_RADIAN = 6371008


def points_on_line(x1: int, y1: int, x2: int, y2: int) -> List[Tuple[int, int]]:
    # Credit: https://stackoverflow.com/questions/25837544
    points = []
    issteep = abs(y2 - y1) > abs(x2 - x1)
    if issteep:
        x1, y1 = y1, x1
        x2, y2 = y2, x2
    rev = False
    if x1 > x2:
        x1, x2 = x2, x1
        y1, y2 = y2, y1
        rev = True
    deltax = x2 - x1
    deltay = abs(y2 - y1)
    error = int(deltax / 2)
    y = y1
    ystep = None
    if y1 < y2:
        ystep = 1
    else:
        ystep = -1
    for x in range(x1, x2 + 1):
        if issteep:
            points.append((y, x))
        else:
            points.append((x, y))
        error -= deltay
        if error < 0:
            y += ystep
            error += deltax
    # Reverse the list if the coordinates were reversed
    if rev:
        points.reverse()
    return points


def get_srtm1_file_path(hgt_name: str):
    paths = SRTM1_DIR.glob(f"**/{hgt_name}.*")
    assert (
        paths
    ), f"Path for HGT name {hgt_name} could not be found in {SRTM1_DIR}. Perhaps there is no file for those coordinates?"
    return paths[0]


def get_srtm3_file_path(hgt_name: str):
    for sub_dir in _HGT_SUBDIRS:
        hgt_path = SRTM3_DIR / sub_dir / f"{hgt_name}.hgt.zip"
        if hgt_path.exists():
            return hgt_path
    assert (
        False
    ), f"Path for HGT name {hgt_name} could not be found. Perhaps there is no file for those coordinates?"


SRTM1_DIR = Path(os.environ.get("SRTM1_DIR", ""))
SRTM3_DIR = Path(os.environ.get("SRTM3_DIR", ""))

_HGT_SUBDIRS = (
    "Eurasia",
    "North_America",
    "Africa",
    "Australia",
    "Islands",
    "South_America",
    "",
)


def haversine(lat1: float, lon1: float, lat2: float, lon2: float):
    """Distance between two lat/long using haversine method"""
    lat1 = radians(lat1)
    lon1 = radians(lon1)
    lat2 = radians(lat2)
    lon2 = radians(lon2)

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = (
        sin(dlat / 2) ** 2
        + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
    )
    c = 2 * atan2(sqrt(a), sqrt(1 - a))

    return EARTH_RADIUS * c


def apply_curvature(profile_points: List["ElevationProfilePoint"]) -> List["ElevationProfilePoint"]:
    """ Apply the earths curvature to the given heights

    The points given should approximate to a straight line.

    Data is a list of tuples, where each tuple is lat, long, height.

    Data is returned in the same form
    """
    left_size = ceil(len(profile_points) / 2)
    left_points = profile_points[:left_size]
    right_points = profile_points[left_size:]

    # Find our mid-point
    if len(left_points) == len(right_points):
        # We have an equal number of points on both sides, so
        # start in between the two center points
        start_lat = mean((left_points[-1].latitude, right_points[0].latitude))
        start_long = mean((left_points[-1].longitude, right_points[0].longitude))
    else:
        # We have an odd number of heights, so start from the center one
        start_lat = left_points[-1].latitude
        start_long = left_points[-1].longitude

    # We we imagine we are walking, and rather then follow the curvature of the
    # earth, we walk out straight on an imaginary plank, seeing the earth slowly
    # fall away below us (because, y'know, it's curved).
    # We then imagine a right angled triangle. The adjacent is the line from
    # our starting position to the center of the earth. The opposite side
    # is the distance we have walked, and the hypotenuse is the distance
    # from our current position to the center of the earth.
    # Given we're now above the earth on our imaginary plank, the hypotenuse
    # should be greater than the earth's radius by however high up we are.
    # If we therefore take away the earths radius, we'll get our altitude.
    height_deltas = []
    for profile_point in profile_points:
        distance = haversine(start_lat, start_long, profile_point.latitude, profile_point.longitude)
        distance_in_radians = distance / METERS_PER_RADIAN
        earth_drop = EARTH_RADIUS / cos(distance_in_radians) - EARTH_RADIUS
        height_deltas.append(earth_drop)

    # Now we have all our altitudes, subtract each one from the heights we were given,
    # thereby reducing all the heights to account for the curvature of the earth.
    adjusted = []
    for profile_point, height_delta in zip(profile_points, height_deltas):
        adjusted.append(profile_point._replace(elevation=profile_point.elevation - height_delta))

    return adjusted


class StraightLineEquation(NamedTuple):
    gradient: float
    c: float

    def y(self, x):
        return self.gradient * x + self.c

    @classmethod
    def from_points(cls, x1, y1, x2, y2) -> "StraightLineEquation":
        gradient = (y2 - y1) / (x2 - x1)
        c = y1 - (gradient * x1)
        return cls(gradient, c)


class ElevationProfilePoint(NamedTuple):
    latitude: float
    longitude: float
    elevation: float
    distance: float


def get_clearances(
    elevation_profile: List[ElevationProfilePoint],
    start_elevation_offset: float = 0,
    end_elevation_offset: float = 0,
) -> List[float]:
    """Get a list of clearances between the line-of-sight and ground level"""
    start = elevation_profile[0]
    end = elevation_profile[-1]

    if start == end:
        return []

    straight_line = StraightLineEquation.from_points(
        x1=start.distance,
        y1=start.elevation + start_elevation_offset,
        x2=end.distance,
        y2=end.elevation + end_elevation_offset,
    )

    clearances = []
    for point in elevation_profile:
        line_of_sight_elevation = straight_line.y(point.distance)
        clearances.append(line_of_sight_elevation - point.elevation)

    return clearances



from pathlib import Path
from typing import Tuple, Callable
from zipfile import ZipFile

from srtm.utilities import get_srtm3_file_path, get_srtm1_file_path
from srtm.base_coordinates import RasterBaseCoordinates


class HeightMap:
    """Provides access to a single SRTM HGT file

    Data will be lazy-loaded on first access
    """

    raster: bytes = None
    base_coordinates: RasterBaseCoordinates
    file_path_fn: Callable = None
    expected_values = 1442401
    values_per_row = 1201

    def __init__(self, path: Path, base_coordinates: RasterBaseCoordinates = None):
        self.path = path
        self.base_coordinates = (
            base_coordinates or RasterBaseCoordinates.from_file_path(path)
        )

        # We subtract one as each row overlaps the neighbouring raster by 1 pixel
        self.pixel_width = 1 / (self.values_per_row - 1)

    @classmethod
    def from_base_coordinates(cls, base_coordinates: RasterBaseCoordinates):
        return cls(
            path=cls.file_path_fn(base_coordinates.file_name),
            base_coordinates=base_coordinates,
        )

    def ensure_loaded(self, force=False):
        """Ensure the file has been loaded from disk"""
        if not force and self.raster is not None:
            return

        if ".zip" in self.path.suffixes:
            zipped_files = ZipFile(self.path).namelist()
            zipped_files = [name for name in zipped_files if ".hgt" in name]
            assert len(zipped_files) == 1, (
                f"ZIP at {self.path} contains the wrong number of hgt files "
                f"({len(zipped_files)}!=1). Contains {zipped_files}"
            )
            self.raster = ZipFile(self.path).read(zipped_files[0])
        else:
            self.raster = self.path.read_bytes()

        self.validate()

    def validate(self):
        """Perform sanity checks"""
        expected_bytes = self.expected_values * 2
        assert len(self.raster) == expected_bytes, (
            f"Unexpected number of bytes found in {self.path}. "
            f"Expected {expected_bytes}, found {len(self.raster)}"
        )

    def get_altitude_for_pixel(self, x, y) -> int:
        """Get the height at the given pixel

        Will trigger loading of data
        """
        self.ensure_loaded()
        # Get the 1-indexed pixel number
        pixel_number = x + (y - 1) * self.values_per_row
        # Convert it ro be 0-indexed
        pixel_number -= 1

        byte_number = pixel_number * 2
        return int.from_bytes(
            self.raster[byte_number : byte_number + 2], byteorder="big"
        )

    def get_altitude_for_latitude_and_longitude(
        self, latitude: float, longitude: float
    ) -> int:
        """Get the height at the given lat/lng"""
        x, y = self._latitude_and_longitude_to_coordinates(latitude, longitude)
        return self.get_altitude_for_pixel(x, y)

    def _latitude_and_longitude_to_coordinates(
        self, latitude: float, longitude: float
    ) -> Tuple[int, int]:
        """Convert the given lat/long into x/y coordinates for this SRTM data"""
        origin_latitude = self.base_coordinates.latitude + 1
        origin_longitude = self.base_coordinates.longitude
        latitude_offset = origin_latitude - latitude
        longitude_offset = longitude - origin_longitude

        if latitude_offset > 1 or latitude_offset < 0:
            raise ValueError(
                f"Latitude {latitude} with offset {latitude_offset} is not within "
                f"this heightmap of base coordinates {self.base_coordinates}"
            )
        if longitude_offset > 1 or longitude_offset < 0:
            raise ValueError(
                f"Longitude {longitude} with offset {longitude_offset} is not within "
                f"this heightmap of base coordinates {self.base_coordinates}"
            )

        # Add one because pixels are 1-indexed
        x = round(longitude_offset / self.pixel_width) + 1
        y = round(latitude_offset / self.pixel_width) + 1

        return x, y


class Srtm1HeightMap(HeightMap):
    """Provides access to a single SRTM HGT file

    Data will be lazy-loaded on first access
    """

    raster: bytes = None
    base_coordinates: RasterBaseCoordinates

    expected_values = 12967201
    values_per_row = 3601
    file_path_fn = get_srtm1_file_path


class Srtm3HeightMap(HeightMap):
    """Provides access to a single SRTM HGT file

    Data will be lazy-loaded on first access
    """

    raster: bytes = None
    base_coordinates: RasterBaseCoordinates
    expected_values = 1442401
    values_per_row = 1201
    file_path_fn = get_srtm3_file_path



from pathlib import Path
from typing import Dict, Type, List, Generator, Tuple

from srtm.base_coordinates import RasterBaseCoordinates
from srtm.exceptions import NoHeightMapDataException
from srtm.utilities import (
    points_on_line,
    SRTM3_DIR,
    SRTM1_DIR,
    apply_curvature,
    haversine,
    ElevationProfilePoint,
)
from srtm.height_maps import HeightMap, Srtm3HeightMap, Srtm1HeightMap


class HeightMapCollection:
    """ Provides access to data across all SRTM files

    This will lazy load data as needed
    """

    height_maps: Dict[RasterBaseCoordinates, HeightMap]
    height_map_class: Type[HeightMap] = None
    hgt_dir: Path = None

    def __init__(self, auto_build_index=True, hgt_dir: Path=None):
        self.height_maps = {}
        if auto_build_index:
            self.build_file_index()

        if hgt_dir is not None:
            self.hgt_dir = hgt_dir

        assert (
            self.height_map_class
        ), "Error, use Srtm3HeightMapCollection or Srtm1HeightMapCollection"

        assert (
            self.hgt_dir
        ), "No HGT direction set. Do you need to set the SRTM1_DIR or SRTM3_DIR environment variables?"

    def build_file_index(self):
        """Load an index of all available files

        This reads file names, but does not load the contained data.
        This is lazy-loaded on demand
        """
        self.height_maps = {}
        for hgt_path in self.hgt_dir.glob("**/*.hgt*"):
            hgt_name = hgt_path.name.split(".")[0]
            self.height_maps[
                RasterBaseCoordinates.from_file_name(hgt_name)
            ] = self.height_map_class(path=hgt_path)

    def get_height_map_for_latitude_and_longitude(
        self, latitude: float, longitude: float
    ) -> HeightMap:
        """Get the HeightMap for the given latitude and longitude"""
        base = RasterBaseCoordinates.from_float(latitude, longitude)
        try:
            return self.height_maps[base]
        except KeyError:
            raise NoHeightMapDataException(
                f"Height map for {base} not found. Have you called "
                f"build_file_index() on your heightmap collection?"
            )

    def get_altitude(self, latitude: float, longitude: float) -> int:
        """Get the height of the given latitude and longitude"""
        height_map = self.get_height_map_for_latitude_and_longitude(latitude, longitude)
        return height_map.get_altitude_for_latitude_and_longitude(latitude, longitude)

    def load_area(self, corner1: RasterBaseCoordinates, corner2: RasterBaseCoordinates):
        """Pre-load a specific area of height maps"""
        min_latitude = min(corner1.latitude, corner2.latitude)
        max_latitude = max(corner1.latitude, corner2.latitude)
        min_longitude = min(corner1.longitude, corner2.longitude)
        max_longitude = max(corner1.longitude, corner2.longitude)

        for height_map in self.height_maps.values():
            ok_latitude = (
                min_latitude <= height_map.base_coordinates.latitude <= max_latitude
            )
            ok_longitude = (
                min_longitude <= height_map.base_coordinates.longitude <= max_longitude
            )
            if ok_latitude and ok_longitude:
                height_map.ensure_loaded()

    def get_elevation_profile(
        self,
        start_latitude: float,
        start_longitude: float,
        end_latitude: float,
        end_longitude: float,
        apply_earth_curvature=True,
    ) -> List[ElevationProfilePoint]:
        """Get the elevation profile between the two points given"""
        values_per_degree = self.height_map_class.values_per_row

        def to_int(lat_lng: float) -> int:
            return round(lat_lng * values_per_degree)

        def to_float(lat_lng_int: int) -> float:
            return lat_lng_int / values_per_degree

        points = points_on_line(
            x1=to_int(start_latitude),
            y1=to_int(start_longitude),
            x2=to_int(end_latitude),
            y2=to_int(end_longitude),
        )
        converted_points = [(to_float(x), to_float(y)) for x, y in points]

        elevations = []
        for latitude, longitude in converted_points:
            elevations.append(
                (latitude, longitude, self.get_altitude(latitude, longitude))
            )

        elevation_points = []
        for latitude, longitude, elevation in elevations:
            elevation_points.append(
                ElevationProfilePoint(
                    latitude,
                    longitude,
                    elevation,
                    haversine(start_latitude, start_longitude, latitude, longitude),
                )
            )

        if apply_earth_curvature:
            elevation_points = apply_curvature(elevation_points)

        return elevation_points

    def get_points(self, min_latitude, min_longitude, max_latitude, max_longitude) -> Generator[Tuple[float, float], None, None]:
        assert min_latitude < max_latitude
        assert min_longitude < max_longitude

        # TODO: Make the pixel width more easily accessible
        step = 1 / (self.height_map_class.values_per_row - 1)
        min_latitude = round(min_latitude / step) * step
        min_longitude = round(min_longitude / step) * step

        latitude = min_latitude
        longitude = min_longitude

        while True:
            if latitude > max_latitude:
                latitude = min_latitude
                longitude += step

                if longitude > max_longitude:
                    break

            yield latitude, longitude
            latitude += step

class Srtm3HeightMapCollection(HeightMapCollection):
    height_map_class = Srtm3HeightMap
    hgt_dir = SRTM3_DIR


class Srtm1HeightMapCollection(HeightMapCollection):
    height_map_class = Srtm1HeightMap
    hgt_dir = SRTM1_DIR



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['srtm']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'python-srtm',
    'version': '0.4.0',
    'description': '',
    'long_description': '# NASA SRTM altitude data parsing in Python\n\nProvides an API onto SRTM `.hgt` or `.hgt.zip` files.\n\nRequires Python 3.8, **may** work with Python 3.6 & 3.7.\n\n## Installation\n\n```\npip install python-srtm\n\nexport SRTM1_DIR=/path/to/srtm1/\nexport SRTM3_DIR=/path/to/srtm3/\n```\n\n## Use\n\nYou can access either SRTM1 or SRTM3 data. SRTM 1, for example:\n\n```python\n# SRTM1 - 30m resolution\n>>> from srtm import Srtm1HeightMapCollection\n>>> srtm1_data = Srtm1HeightMapCollection()\n>>> srtm1_data.get_altitude(latitude=40.123, longitude=-7.456)\n615\n>>> Srtm1HeightMapCollection().get_elevation_profile(40.123, -7.456, 40.129, -7.460)\n[615, 620, 618, 620, 616, 603, 593, 582, 575, 579, 580, 589, 589, 581, 565, 553, 545, 541, 534, 533, 529, 520, 514]\n```\n\nOr SRTM3:\n\n```python\n# SRTM3 - 90m resolution\n>>> from srtm import Srtm3HeightMapCollection\n>>> srtm3_data = Srtm3HeightMapCollection()\n>>> srtm3_data.get_altitude(latitude=40.123, longitude=-7.456)\n608\n>>> Srtm3HeightMapCollection().get_elevation_profile(40.123, -7.456, 40.129, -7.460)\n[626, 616, 585, 593, 577, 548, 528, 514]\n```\n\n## Profiling\n\n```python\nimport cProfile\ncProfile.run(\'function_to_profile()\', filename=\'output.cprof\')\n```\n\n```bash\nbrew install qcachegrind\npip install pyprof2calltree\npyprof2calltree -k -i /pythonprofiling/profiler/first_iteration.cprof\n```\n\n## Release process\n\nFor internal reference:\n\n```\n# Run the tests\npytest\n\n# Update the setup.py\ndephell convert\nblack setup.py\n\n# Ensure poetry.lock is up to date\npoetry lock\n\nexport VERSION="VERSION HERE"\n\n# Version bump\npoetry version $VERSION\n\n\n# Commit\ngit add .\ngit commit -m "Releasing version $VERSION"\n\n# Tagging and branching\ngit tag "v$VERSION"\ngit branch "v$VERSION"\ngit push origin \\\n    refs/tags/"v$VERSION" \\\n    refs/heads/"v$VERSION" \\\n    master\n\npoetry publish --build\n```\n',
    'author': 'Adam Charnock',
    'author_email': 'adam@adamcharnock.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/adamcharnock/python-srtm',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
